import numpy as np
from autolens.imaging import mask
import pytest
from autolens import exc


@pytest.fixture(name="msk")
def make_mask():
    return mask.Mask(np.array([[True, False, True],
                               [False, False, False],
                               [True, False, True]]))


@pytest.fixture(name="centre_mask")
def make_centre_mask():
    return mask.Mask(np.array([[True, True, True],
                               [True, False, True],
                               [True, True, True]]))


@pytest.fixture(name="sub_coordinate_grid")
def make_sub_coordinate_grid(msk):
    return mask.SubGrid.from_mask(msk)


class TestMask(object):
    class TestConstructor(object):

        def test__simple_array_in(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1)

            assert (msk == np.array([[True, True, True],
                                     [True, False, True],
                                     [True, True, True]])).all()
            assert msk.pixel_scale == 1.0
            assert msk.central_pixel_coordinates == (1.0, 1.0)
            assert msk.shape == (3, 3)
            assert msk.shape_arc_seconds == (3.0, 3.0)

        def test__rectangular_array_in(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1)

            assert (msk == np.array([[True, True, True, True],
                                     [True, False, False, True],
                                     [True, True, True, True]])).all()
            assert msk.pixel_scale == 1.0
            assert msk.central_pixel_coordinates == (1.0, 1.5)
            assert msk.shape == (3, 4)
            assert msk.shape_arc_seconds == (3.0, 4.0)

    class TestCircular(object):

        def test__input_big_mask__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1.0, radius_mask=5)

            assert msk.shape == (3, 3)
            assert (msk == np.array([[False, False, False],
                                     [False, False, False],
                                     [False, False, False]])).all()

        def test__3x3_mask_input_radius_small__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=0.5)
            assert (msk == np.array([[True, True, True],
                                     [True, False, True],
                                     [True, True, True]])).all()

        def test__3x3_mask_input_radius_medium__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=1)

            assert (msk == np.array([[True, False, True],
                                     [False, False, False],
                                     [True, False, True]])).all()

        def test__3x3_mask_input_radius_large__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=3)

            assert (msk == np.array([[False, False, False],
                                     [False, False, False],
                                     [False, False, False]])).all()

        def test__4x3_mask_input_radius_small__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 3), pixel_scale=1, radius_mask=0.5)

            assert (msk == np.array([[True, True, True],
                                     [True, False, True],
                                     [True, False, True],
                                     [True, True, True]])).all()

        def test__4x3_mask_input_radius_medium__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 3), pixel_scale=1, radius_mask=1.50001)

            assert (msk == np.array([[True, False, True],
                                     [False, False, False],
                                     [False, False, False],
                                     [True, False, True]])).all()

        def test__4x3_mask_input_radius_large__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 3), pixel_scale=1, radius_mask=3)

            assert (msk == np.array([[False, False, False],
                                     [False, False, False],
                                     [False, False, False],
                                     [False, False, False]])).all()

        def test__4x4_mask_input_radius_small__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 4), pixel_scale=1, radius_mask=0.72)

            assert (msk == np.array([[True, True, True, True],
                                     [True, False, False, True],
                                     [True, False, False, True],
                                     [True, True, True, True]])).all()

        def test__4x4_mask_input_radius_medium__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 4), pixel_scale=1, radius_mask=1.7)

            assert (msk == np.array([[True, False, False, True],
                                     [False, False, False, False],
                                     [False, False, False, False],
                                     [True, False, False, True]])).all()

        def test__4x4_mask_input_radius_large__mask(self):
            msk = mask.Mask.circular(shape_arc_seconds=(4, 4), pixel_scale=1, radius_mask=3)

            assert (msk == np.array([[False, False, False, False],
                                     [False, False, False, False],
                                     [False, False, False, False],
                                     [False, False, False, False]])).all()

        def test__centre_shift__simple_shift_back(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=0.5, centre=(-1, 0))

            assert msk.shape == (3, 3)
            assert (msk == np.array([[True, False, True],
                                     [True, True, True],
                                     [True, True, True]])).all()

        def test__centre_shift__simple_shift_forward(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=0.5, centre=(0, 1))

            assert msk.shape == (3, 3)
            assert (msk == np.array([[True, True, True],
                                     [True, True, False],
                                     [True, True, True]])).all()

        def test__centre_shift__diagonal_shift(self):
            msk = mask.Mask.circular(shape_arc_seconds=(3, 3), pixel_scale=1, radius_mask=0.5, centre=(1, 1))

            assert (msk == np.array([[True, True, True],
                                     [True, True, True],
                                     [True, True, False]])).all()

    class TestAnnular(object):

        def test__3x3_mask_inner_radius_zero_outer_radius_small__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(3, 3), pixel_scale=1, inner_radius=0,
                                    outer_radius=0.5)

            assert (msk == np.array([[True, True, True],
                                     [True, False, True],
                                     [True, True, True]])).all()

        def test__3x3_mask_inner_radius_small_outer_radius_large__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(3, 3), pixel_scale=1, inner_radius=0.5,
                                    outer_radius=3)

            assert (msk == np.array([[False, False, False],
                                     [False, True, False],
                                     [False, False, False]])).all()

        def test__4x4_mask_inner_radius_small_outer_radius_medium__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(4, 3), pixel_scale=1, inner_radius=0.51,
                                    outer_radius=1.51)

            assert (msk == np.array([[True, False, True],
                                     [False, True, False],
                                     [False, True, False],
                                     [True, False, True]])).all()

        def test__4x3_mask_inner_radius_medium_outer_radius_large__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(4, 3), pixel_scale=1, inner_radius=1.51,
                                    outer_radius=3)

            assert (msk == np.array([[False, True, False],
                                     [True, True, True],
                                     [True, True, True],
                                     [False, True, False]])).all()

        def test__3x3_mask_inner_radius_small_outer_radius_medium__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(4, 4), pixel_scale=1, inner_radius=0.81,
                                    outer_radius=2)

            assert (msk == np.array([[True, False, False, True],
                                     [False, True, True, False],
                                     [False, True, True, False],
                                     [True, False, False, True]])).all()

        def test__4x4_mask_inner_radius_medium_outer_radius_large__mask(self):
            msk = mask.Mask.annular(shape_arc_seconds=(4, 4), pixel_scale=1, inner_radius=1.71,
                                    outer_radius=3)

            assert (msk == np.array([[False, True, True, False],
                                     [True, True, True, True],
                                     [True, True, True, True],
                                     [False, True, True, False]])).all()

        def test__centre_shift__simple_shift_back(self):
            msk = mask.Mask.annular(shape_arc_seconds=(3, 3), pixel_scale=1, inner_radius=0.5,
                                    outer_radius=3, centre=(-1.0, 0.0))

            assert msk.shape == (3, 3)
            assert (msk == np.array([[False, True, False],
                                     [False, False, False],
                                     [False, False, False]])).all()

        def test__centre_shift__simple_shift_forward(self):
            msk = mask.Mask.annular(shape_arc_seconds=(3, 3), pixel_scale=1, inner_radius=0.5,
                                    outer_radius=3, centre=(0.0, 1.0))

            assert msk.shape == (3, 3)
            assert (msk == np.array([[False, False, False],
                                     [False, False, True],
                                     [False, False, False]])).all()

        def test__centre_shift__diagonal_shift(self):
            msk = mask.Mask.annular(shape_arc_seconds=(3, 3), pixel_scale=1, inner_radius=0.5,
                                    outer_radius=3, centre=(1.0, 1.0))

            assert msk.shape == (3, 3)
            assert (msk == np.array([[False, False, False],
                                     [False, False, False],
                                     [False, False, True]])).all()

    class TestUnmasked(object):

        def test__3x3__input__all_are_false(self):
            msk = mask.Mask.unmasked(shape_arc_seconds=(3, 3), pixel_scale=1)

            assert msk.shape == (3, 3)
            assert (msk == np.array([[False, False, False],
                                     [False, False, False],
                                     [False, False, False]])).all()

        def test__3x2__input__all_are_false(self):
            msk = mask.Mask.unmasked(shape_arc_seconds=(1.5, 1.0), pixel_scale=0.5)

            assert msk.shape == (3, 2)
            assert (msk == np.array([[False, False],
                                     [False, False],
                                     [False, False]])).all()

        def test__5x5__input__all_are_false(self):
            msk = mask.Mask.unmasked(shape_arc_seconds=(5, 5), pixel_scale=1)

            assert msk.shape == (5, 5)
            assert (msk == np.array([[False, False, False, False, False],
                                     [False, False, False, False, False],
                                     [False, False, False, False, False],
                                     [False, False, False, False, False],
                                     [False, False, False, False, False]])).all()

    class TestForSimulate(object):

        def test__3x3_image__3x3_psf_size__5x5_image_made_border_all_masked(self):

            msk = mask.Mask.for_simulate(shape_arc_seconds=(3, 3), pixel_scale=1, psf_size=(3, 3))

            assert (msk == np.array([[True, True, True, True, True],
                                     [True, False, False, False, True],
                                     [True, False, False, False, True],
                                     [True, False, False, False, True],
                                     [True, True, True, True, True]])).all()

        def test__3x3_image__5x5_psf_size__7x7_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(3, 3), pixel_scale=1, psf_size=(5, 5))

            assert (msk == np.array([[True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True]])).all()

        def test__3x3_image__7x7_psf_size__7x7_image_made_border_all_masked(self):

            msk = mask.Mask.for_simulate(shape_arc_seconds=(3, 3), pixel_scale=1, psf_size=(7, 7))

            assert (msk == np.array([[True, True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True, True],
                                     [True, True, True, False, False, False, True, True, True],
                                     [True, True, True, False, False, False, True, True, True],
                                     [True, True, True, False, False, False, True, True, True],
                                     [True, True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True, True]])).all()

        def test__4x3_image__3x3_psf_size__6x5_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(4, 3), pixel_scale=1, psf_size=(3, 3))

            assert (msk == np.array([[True, True, True, True, True],
                                     [True, False, False, False, True],
                                     [True, False, False, False, True],
                                     [True, False, False, False, True],
                                     [True, False, False, False, True],
                                     [True, True, True, True, True]])).all()

        def test__4x3_image__5x5_psf_size__8x7_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(4, 3), pixel_scale=1, psf_size=(5, 5))

            assert (msk == np.array([[True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, False, False, False, True, True],
                                     [True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True]])).all()

        def test__3x4_image__3x3_psf_size__5x6_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(3, 4), pixel_scale=1, psf_size=(3, 3))

            assert (msk == np.array([[True, True, True, True, True, True],
                                     [True, False, False, False, False, True],
                                     [True, False, False, False, False, True],
                                     [True, False, False, False, False, True],
                                     [True, True, True, True, True, True]])).all()

        def test__3x4_image__5x5_psf_size__7x8_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(3, 4), pixel_scale=1, psf_size=(5, 5))

            assert (msk == np.array([[True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True]])).all()

        def test__4x4_image__3x3_psf_size__6x6_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(4, 4), pixel_scale=1, psf_size=(3, 3))

            assert (msk == np.array([[True, True, True, True, True, True],
                                     [True, False, False, False, False, True],
                                     [True, False, False, False, False, True],
                                     [True, False, False, False, False, True],
                                     [True, False, False, False, False, True],
                                     [True, True, True, True, True, True]])).all()

        def test__4x4_image__5x5_psf_size__8x8_image_made_border_all_masked(self):
            msk = mask.Mask.for_simulate(shape_arc_seconds=(4, 4), pixel_scale=1, psf_size=(5, 5))

            assert (msk == np.array([[True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, False, False, False, False, True, True],
                                     [True, True, True, True, True, True, True, True],
                                     [True, True, True, True, True, True, True, True]])).all()

        def test__non_square_psf_shape__raises_error(self):
            with pytest.raises(exc.KernelException):
                mask.Mask.for_simulate(shape_arc_seconds=(4, 4), pixel_scale=1, psf_size=(3, 5))

    class TestComputeGridCoordsImage(object):

        def test__setup_3x3_image_one_coordinate(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_grid = msk.coordinate_grid

            assert (image_grid[0] == np.array([0.0, 0.0])).all()

        def test__setup_3x3_image__five_grid(self):
            msk = np.array([[True, False, True],
                            [False, False, False],
                            [True, False, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_grid = msk.coordinate_grid

            assert (image_grid == np.array([[-3., 0.], [0., -3.], [0., 0.], [0., 3.], [3., 0.]])).all()

        def test__setup_4x4_image__ten_grid__new_pixel_scale(self):
            msk = np.array([[True, False, False, True],
                            [False, False, False, True],
                            [True, False, False, True],
                            [False, False, False, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            image_grid = msk.coordinate_grid

            assert (image_grid == np.array(
                [[-1.5, -0.5], [-1.5, 0.5], [-0.5, -1.5], [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5],
                 [1.5, -1.5], [1.5, -0.5], [1.5, 0.5]])).all()

        def test__setup_3x4_image__six_grid(self):
            msk = np.array([[True, False, True, True],
                            [False, False, False, True],
                            [True, False, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_grid = msk.coordinate_grid

            assert (image_grid == np.array(
                [[-3., -1.5], [0., -4.5], [0., -1.5], [0., 1.5], [3., -1.5], [3., 4.5]])).all()

    class TestComputeGridCoordsImageSub(object):

        def test__3x3_mask_with_one_pixel__2x2_sub_grid__grid(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            assert (image_sub_grid == np.array([[[-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5]]])).all()

        def test__3x3_mask_with_row_of_pixels__2x2_sub_grid__grid(self):
            msk = np.array([[True, True, True],
                            [False, False, False],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            assert (image_sub_grid == np.array([[-0.5, -3.5], [-0.5, -2.5], [0.5, -3.5], [0.5, -2.5],
                                                [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5],
                                                [-0.5, 2.5], [-0.5, 3.5], [0.5, 2.5], [0.5, 3.5]])).all()

        def test__3x3_mask_with_row_and_column_of_pixels__2x2_sub_grid__grid(self):
            msk = np.array([[True, True, False],
                            [False, False, False],
                            [True, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            assert (image_sub_grid == np.array([[-3.5, 2.5], [-3.5, 3.5], [-2.5, 2.5], [-2.5, 3.5],
                                                [-0.5, -3.5], [-0.5, -2.5], [0.5, -3.5], [0.5, -2.5],
                                                [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5],
                                                [-0.5, 2.5], [-0.5, 3.5], [0.5, 2.5], [0.5, 3.5],
                                                [2.5, 2.5], [2.5, 3.5], [3.5, 2.5], [3.5, 3.5]])).all()

        def test__3x3_mask_with_row_and_column_of_pixels__2x2_sub_grid__different_pixel_scale(self):
            msk = np.array([[True, True, False],
                            [False, False, False],
                            [True, True, False]])

            msk = mask.Mask(msk, pixel_scale=0.3)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            image_sub_grid = np.round(image_sub_grid, decimals=2)

            np.testing.assert_almost_equal(image_sub_grid,
                                           np.array([[-0.35, 0.25], [-0.35, 0.35], [-0.25, 0.25], [-0.25, 0.35],
                                                     [-0.05, -0.35], [-0.05, -0.25], [0.05, -0.35], [0.05, -0.25],
                                                     [-0.05, -0.05], [-0.05, 0.05], [0.05, -0.05], [0.05, 0.05],
                                                     [-0.05, 0.25], [-0.05, 0.35], [0.05, 0.25], [0.05, 0.35],
                                                     [0.25, 0.25], [0.25, 0.35], [0.35, 0.25], [0.35, 0.35]]))

        def test__3x3_mask_with_one_pixel__3x3_sub_grid__grid(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 3)

            assert (image_sub_grid == np.array([[[-0.75, -0.75], [-0.75, 0.], [-0.75, 0.75], [0., -0.75], [0., 0.],
                                                 [0., 0.75], [0.75, -0.75], [0.75, 0.], [0.75, 0.75]]])).all()

        def test__3x3_mask_with_one_row__3x3_sub_grid__grid(self):
            msk = np.array([[True, True, False],
                            [True, False, True],
                            [True, True, False]])

            msk = mask.Mask(msk, pixel_scale=2.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 3)

            assert (image_sub_grid == np.array([[-2.5, 1.5], [-2.5, 2.], [-2.5, 2.5], [-2., 1.5], [-2., 2.],
                                                [-2., 2.5], [-1.5, 1.5], [-1.5, 2.], [-1.5, 2.5],
                                                [-0.5, -0.5], [-0.5, 0.], [-0.5, 0.5], [0., -0.5], [0., 0.], [0., 0.5],
                                                [0.5, -0.5], [0.5, 0.], [0.5, 0.5],
                                                [1.5, 1.5], [1.5, 2.], [1.5, 2.5], [2., 1.5], [2., 2.], [2., 2.5],
                                                [2.5, 1.5], [2.5, 2.], [2.5, 2.5]])).all()

        def test__4x4_mask_with_one_pixel__4x4_sub_grid__grid(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, False, False, True],
                            [True, True, True, False]])

            msk = mask.Mask(msk, pixel_scale=2.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 4)

            image_sub_grid = np.round(image_sub_grid, decimals=1)

            assert (image_sub_grid == np.array([[-1.6, -1.6], [-1.6, -1.2], [-1.6, -0.8], [-1.6, -0.4], [-1.2, -1.6],
                                                [-1.2, -1.2], [-1.2, -0.8], [-1.2, -0.4], [-0.8, -1.6], [-0.8, -1.2],
                                                [-0.8, -0.8], [-0.8, -0.4], [-0.4, -1.6], [-0.4, -1.2], [-0.4, -0.8],
                                                [-0.4, -0.4],
                                                [-1.6, 0.4], [-1.6, 0.8], [-1.6, 1.2], [-1.6, 1.6], [-1.2, 0.4],
                                                [-1.2, 0.8], [-1.2, 1.2], [-1.2, 1.6], [-0.8, 0.4], [-0.8, 0.8],
                                                [-0.8, 1.2], [-0.8, 1.6], [-0.4, 0.4], [-0.4, 0.8], [-0.4, 1.2],
                                                [-0.4, 1.6],
                                                [0.4, -1.6], [0.4, -1.2], [0.4, -0.8], [0.4, -0.4], [0.8, -1.6],
                                                [0.8, -1.2], [0.8, -0.8], [0.8, -0.4], [1.2, -1.6], [1.2, -1.2],
                                                [1.2, -0.8], [1.2, -0.4], [1.6, -1.6], [1.6, -1.2], [1.6, -0.8],
                                                [1.6, -0.4],
                                                [0.4, 0.4], [0.4, 0.8], [0.4, 1.2], [0.4, 1.6], [0.8, 0.4], [0.8, 0.8],
                                                [0.8, 1.2], [0.8, 1.6], [1.2, 0.4], [1.2, 0.8], [1.2, 1.2], [1.2, 1.6],
                                                [1.6, 0.4], [1.6, 0.8], [1.6, 1.2], [1.6, 1.6],
                                                [2.4, 2.4], [2.4, 2.8], [2.4, 3.2], [2.4, 3.6], [2.8, 2.4], [2.8, 2.8],
                                                [2.8, 3.2], [2.8, 3.6], [3.2, 2.4], [3.2, 2.8], [3.2, 3.2], [3.2, 3.6],
                                                [3.6, 2.4], [3.6, 2.8], [3.6, 3.2], [3.6, 3.6]])).all()

        def test__4x3_mask_with_one_pixel__2x2_sub_grid__grid(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, False, False],
                            [False, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            assert (image_sub_grid == np.array(
                [[-2., -0.5], [-2., 0.5], [-1., -0.5], [-1., 0.5], [1., -0.5], [1., 0.5], [2., -0.5], [2., 0.5],
                 [1., 2.5], [1., 3.5], [2., 2.5], [2., 3.5], [4., -3.5], [4., -2.5], [5., -3.5], [5., -2.5]])).all()

        def test__3x4_mask_with_one_pixel__2x2_sub_grid__grid(self):
            msk = np.array([[True, True, True, False],
                            [True, False, False, True],
                            [False, True, False, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            image_sub_grid = mask.SubGrid.from_mask(msk, 2)

            assert (image_sub_grid == np.array(
                [[-3.5, 4.], [-3.5, 5.], [-2.5, 4.], [-2.5, 5.], [-0.5, -2.], [-0.5, -1.], [0.5, -2.], [0.5, -1.],
                 [-0.5, 1.], [-0.5, 2.], [0.5, 1.], [0.5, 2.], [2.5, -5.], [2.5, -4.], [3.5, -5.], [3.5, -4.],
                 [2.5, 1.], [2.5, 2.], [3.5, 1.], [3.5, 2.]])).all()

    class TestComputeGridCoordsBlurring(object):

        def test__3x3_blurring_mask_correct_grid(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_grid = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3)).coordinate_grid

            assert (blurring_grid == np.array(
                [[-3., -3.], [-3., 0.], [-3., 3.], [0., -3.], [0., 3.], [3., -3.], [3., 0.], [3., 3.]])).all()

        def test__3x5_blurring_mask_correct_grid(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            # Blurring mask

            # [[True, True, True, True, True, True, True],
            # [True, True, False, False, False, True, True],
            # [True, True, False, False, False, True, True],
            # [True, True, False, True, False, True, True],
            # [True, True, False, False, False, True, True],
            # [True, True, False, False, False, True, True],
            # [True, True, True, True, True, True, True]])

            blurring_grid = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 5)).coordinate_grid

            assert (blurring_grid == np.array(
                [[-3., -6.], [-3., -3.], [-3., 0.], [-3., 3.], [-3., 6.], [0., -6.], [0., -3.], [0., 3.], [0., 6.],
                 [3., -6.], [3., -3.], [3., 0.], [3., 3.], [3., 6.]])).all()

        def test__5x3_blurring_mask_correct_grid(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            # Blurring mask

            # [[True, True, True, True, True, True, True],
            #  [True, True, True, True, True, True, True],
            #  [True, False, False, False, False, False, True],
            #  [True, False, False, True, False, False, True],
            #  [True, False, False, False, False, False, True],
            #  [True, True, True, True, True, True, True],
            #  [True, True, True, True, True, True, True]]

            blurring_grid = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 3)).coordinate_grid

            assert (blurring_grid == np.array(
                [[-6., -3.], [-6., 0.], [-6., 3.], [-3., -3.], [-3., 0.], [-3., 3.], [0., -3.], [0., 3.], [3., -3.],
                 [3., 0.], [3., 3.], [6., -3.], [6., 0.], [6., 3.]])).all()

    class TestComputeGridSubtoImage(object):

        def test__3x3_mask_with_1_pixel__2x2_sub_grid__correct_sub_to_image(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            sub_to_image = mask.SubGrid.from_mask(msk, 2).sub_to_image

            assert (sub_to_image == np.array([0, 0, 0, 0])).all()

        def test__3x3_mask_with_row_of_pixels_pixel__2x2_sub_grid__correct_sub_to_image(self):
            msk = np.array([[True, True, True],
                            [False, False, False],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            sub_to_image = mask.SubGrid.from_mask(msk, 2).sub_to_image

            assert (sub_to_image == np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])).all()

        def test__3x3_mask_with_row_of_pixels_pixel__3x3_sub_grid__correct_sub_to_image(self):
            msk = np.array([[True, True, True],
                            [False, False, False],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            sub_to_image = mask.SubGrid.from_mask(msk, 3).sub_to_image

            assert (sub_to_image == np.array([0, 0, 0, 0, 0, 0, 0, 0, 0,
                                              1, 1, 1, 1, 1, 1, 1, 1, 1,
                                              2, 2, 2, 2, 2, 2, 2, 2, 2])).all()

    class TestComputeGridData(object):

        def test__setup_3x3_data(self):
            data = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9]])

            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_data = msk.map_to_1d(data)

            assert (grid_data[0] == np.array([5])).all()

        def test__setup_3x3_data__five_now_in_mask(self):
            data = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9]])

            msk = np.array([[True, False, True],
                            [False, False, False],
                            [True, False, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_data = msk.map_to_1d(data)

            assert (grid_data[0] == np.array([2])).all()
            assert (grid_data[1] == np.array([4])).all()
            assert (grid_data[2] == np.array([5])).all()
            assert (grid_data[3] == np.array([6])).all()
            assert (grid_data[4] == np.array([8])).all()

        def test__setup_3x4_data(self):
            data = np.array([[1, 2, 3, 4],
                             [5, 6, 7, 8],
                             [9, 10, 11, 12]])

            msk = np.array([[True, False, True, True],
                            [False, False, False, True],
                            [True, False, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_data = msk.map_to_1d(data)

            assert (grid_data[0] == np.array([2])).all()
            assert (grid_data[1] == np.array([5])).all()
            assert (grid_data[2] == np.array([6])).all()
            assert (grid_data[3] == np.array([7])).all()
            assert (grid_data[4] == np.array([10])).all()

        def test__setup_4x3_data__five_now_in_mask(self):
            data = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9],
                             [10, 11, 12]])

            msk = np.array([[True, False, True],
                            [False, False, False],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_data = msk.map_to_1d(data)

            assert (grid_data[0] == np.array([2])).all()
            assert (grid_data[1] == np.array([4])).all()
            assert (grid_data[2] == np.array([5])).all()
            assert (grid_data[3] == np.array([6])).all()
            assert (grid_data[4] == np.array([8])).all()

    class TestComputeBlurringMask(object):

        def test__size__3x3_small_mask(self):

            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[False, False, False],
                                               [False, True, False],
                                               [False, False, False]])).all()

        def test__size__3x3__large_mask(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[True, True, True, True, True, True, True],
                                               [True, True, True, True, True, True, True],
                                               [True, True, False, False, False, True, True],
                                               [True, True, False, True, False, True, True],
                                               [True, True, False, False, False, True, True],
                                               [True, True, True, True, True, True, True],
                                               [True, True, True, True, True, True, True]])).all()

        def test__size__5x5__large_mask(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 5))

            assert (blurring_mask == np.array([[True, True, True, True, True, True, True],
                                               [True, False, False, False, False, False, True],
                                               [True, False, False, False, False, False, True],
                                               [True, False, False, True, False, False, True],
                                               [True, False, False, False, False, False, True],
                                               [True, False, False, False, False, False, True],
                                               [True, True, True, True, True, True, True]])).all()

        def test__size__5x3__large_mask(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 3))

            assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True],
                                                        [True, True, True, True, True, True, True],
                                                        [True, False, False, False, False, False, True],
                                                        [True, False, False, True, False, False, True],
                                                        [True, False, False, False, False, False, True],
                                                        [True, True, True, True, True, True, True],
                                                        [True, True, True, True, True, True, True]]))).all()

        def test__size__3x5__large_mask(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 5))

            assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True],
                                                        [True, True, False, False, False, True, True],
                                                        [True, True, False, False, False, True, True],
                                                        [True, True, False, True, False, True, True],
                                                        [True, True, False, False, False, True, True],
                                                        [True, True, False, False, False, True, True],
                                                        [True, True, True, True, True, True, True]]))).all()

        def test__size__3x3__multiple_points(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[False, False, False, True, False, False, False],
                                               [False, True, False, True, False, True, False],
                                               [False, False, False, True, False, False, False],
                                               [True, True, True, True, True, True, True],
                                               [False, False, False, True, False, False, False],
                                               [False, True, False, True, False, True, False],
                                               [False, False, False, True, False, False, False]])).all()

        def test__size__5x5__multiple_points(self):
            msk = np.array([[True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 5))

            assert (blurring_mask == np.array([[False, False, False, False, False, False, False, False, False],
                                               [False, False, False, False, False, False, False, False, False],
                                               [False, False, True, False, False, False, True, False, False],
                                               [False, False, False, False, False, False, False, False, False],
                                               [False, False, False, False, False, False, False, False, False],
                                               [False, False, False, False, False, False, False, False, False],
                                               [False, False, True, False, False, False, True, False, False],
                                               [False, False, False, False, False, False, False, False, False],
                                               [False, False, False, False, False, False, False, False,
                                                False]])).all()

        def test__size__5x3__multiple_points(self):
            msk = np.array([[True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 3))

            assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True, True, True],
                                                        [False, False, False, False, False, False, False, False, False],
                                                        [False, False, True, False, False, False, True, False, False],
                                                        [False, False, False, False, False, False, False, False, False],
                                                        [True, True, True, True, True, True, True, True, True],
                                                        [False, False, False, False, False, False, False, False, False],
                                                        [False, False, True, False, False, False, True, False, False],
                                                        [False, False, False, False, False, False, False, False, False],
                                                        [True, True, True, True, True, True, True, True, True]]))).all()

        def test__size__3x5__multiple_points(self):
            msk = np.array([[True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 5))

            assert (blurring_mask == np.rot90(np.array([[True, False, False, False, True, False, False, False, True],
                                                        [True, False, False, False, True, False, False, False, True],
                                                        [True, False, True, False, True, False, True, False, True],
                                                        [True, False, False, False, True, False, False, False, True],
                                                        [True, False, False, False, True, False, False, False, True],
                                                        [True, False, False, False, True, False, False, False, True],
                                                        [True, False, True, False, True, False, True, False, True],
                                                        [True, False, False, False, True, False, False, False, True],
                                                        [True, False, False, False, True, False, False, False,
                                                         True]]))).all()

        def test__size__3x3__even_sized_image(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True],
                                               [False, True, False, True, False, True, False, True],
                                               [False, False, False, True, False, False, False, True],
                                               [True, True, True, True, True, True, True, True],
                                               [False, False, False, True, False, False, False, True],
                                               [False, True, False, True, False, True, False, True],
                                               [False, False, False, True, False, False, False, True],
                                               [True, True, True, True, True, True, True, True]])).all()

        def test__size__5x5__even_sized_image(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 5))

            assert (blurring_mask == np.array([[True, True, True, True, True, True, True, True],
                                               [True, True, True, True, True, True, True, True],
                                               [True, True, True, True, True, True, True, True],
                                               [True, True, True, False, False, False, False, False],
                                               [True, True, True, False, False, False, False, False],
                                               [True, True, True, False, False, True, False, False],
                                               [True, True, True, False, False, False, False, False],
                                               [True, True, True, False, False, False, False, False]])).all()

        def test__size__3x3__rectangular_8x9_image(self):
            msk = np.array([[True, True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True, True],
                                               [False, True, False, True, False, True, False, True, True],
                                               [False, False, False, True, False, False, False, True, True],
                                               [True, True, True, True, True, True, True, True, True],
                                               [False, False, False, True, False, False, False, True, True],
                                               [False, True, False, True, False, True, False, True, True],
                                               [False, False, False, True, False, False, False, True, True],
                                               [True, True, True, True, True, True, True, True, True]])).all()

        def test__size__3x3__rectangular_9x8_image(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            blurring_mask = msk.blurring_mask_for_kernel_shape(kernel_shape=(3, 3))

            assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True],
                                               [False, True, False, True, False, True, False, True],
                                               [False, False, False, True, False, False, False, True],
                                               [True, True, True, True, True, True, True, True],
                                               [False, False, False, True, False, False, False, True],
                                               [False, True, False, True, False, True, False, True],
                                               [False, False, False, True, False, False, False, True],
                                               [True, True, True, True, True, True, True, True],
                                               [True, True, True, True, True, True, True, True]])).all()

        def test__size__5x5__multiple_points__mask_extends_beyond_border_so_raises_mask_exception(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, True, True, True, False, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            with pytest.raises(exc.MaskException):
                msk.blurring_mask_for_kernel_shape(kernel_shape=(5, 5))

    class TestComputeBorderPixels(object):

        def test__7x7_mask_one_central_pixel__is_entire_border(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0])).all()

        def test__7x7_mask_nine_central_pixels__is_border(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 5, 6, 7, 8])).all()

        def test__7x7_mask_rectangle_of_fifteen_central_pixels__is_border(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 5, 6, 8, 9, 11, 12, 13, 14])).all()

        def test__8x7_mask_add_edge_pixels__also_in_border(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15, 16, 17])).all()

        def test__8x7_mask_big_square(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 24, 25, 26, 27, 28, 29])).all()

        def test__7x8_mask_add_edge_pixels__also_in_border(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14])).all()

        def test__7x8_mask_big_square(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_pixels = msk.border_pixel_indices

            assert (border_pixels == np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21, 22, 23, 24])).all()

    class TestComputeBorderSubPixels(object):

        def test__7x7_mask__2x2_sub_grid__nine_central_pixels__is_border(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_sub_pixels = msk.border_sub_pixel_indices(sub_grid_size=2)

            assert (border_sub_pixels == np.array([0, 4, 9, 12, 21, 26, 30, 35])).all()

        def test__7x7_mask__4x4_sub_grid_nine_central_pixels__is_border(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_sub_pixels = msk.border_sub_pixel_indices(sub_grid_size=4)

            assert (border_sub_pixels == np.array([0, 16, 35, 48, 83, 108, 124, 143])).all()

        def test__7x7_mask_rectangle_of_fifteen_central_pixels__is_border(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_sub_pixels = msk.border_sub_pixel_indices(sub_grid_size=2)

            assert (border_sub_pixels == np.array([0, 4, 9, 12, 21, 24, 33, 38, 47, 50, 54, 59])).all()

        def test__8x7_mask_add_edge_pixels__also_in_border(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_sub_pixels = msk.border_sub_pixel_indices(sub_grid_size=2)
            assert (border_sub_pixels == np.array([0, 4, 8, 13, 16, 25, 30, 34, 43, 47, 50, 59, 62, 66, 71])).all()

        def test__7x8_mask_add_edge_pixels__also_in_border(self):

            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, False, True, True, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, False, False, False, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_sub_pixels = msk.border_sub_pixel_indices(sub_grid_size=2)
            assert (border_sub_pixels == np.array([0, 4, 8, 13, 16, 25, 30, 34, 43, 47, 50, 54, 59])).all()

    class TestComputeGridMapperTo2D(object):

        def test__setup_3x3_image_one_pixel(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            mapper_data_to_2d = msk.grid_to_pixel()

            assert (mapper_data_to_2d[0] == np.array([1, 1])).all()

        def test__setup_3x3_image__five_pixels(self):
            msk = np.array([[True, False, True],
                            [False, False, False],
                            [True, False, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            mapper_data_to_2d = msk.grid_to_pixel()

            assert (mapper_data_to_2d[0] == np.array([0, 1])).all()
            assert (mapper_data_to_2d[1] == np.array([1, 0])).all()
            assert (mapper_data_to_2d[2] == np.array([1, 1])).all()
            assert (mapper_data_to_2d[3] == np.array([1, 2])).all()
            assert (mapper_data_to_2d[4] == np.array([2, 1])).all()

        def test__setup_3x4_image__six_pixels(self):
            msk = np.array([[True, False, True, True],
                            [False, False, False, True],
                            [True, False, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            mapper_data_to_2d = msk.grid_to_pixel()

            assert (mapper_data_to_2d[0] == np.array([0, 1])).all()
            assert (mapper_data_to_2d[1] == np.array([1, 0])).all()
            assert (mapper_data_to_2d[2] == np.array([1, 1])).all()
            assert (mapper_data_to_2d[3] == np.array([1, 2])).all()
            assert (mapper_data_to_2d[4] == np.array([2, 1])).all()
            assert (mapper_data_to_2d[5] == np.array([2, 3])).all()

        def test__setup_4x3_image__six_pixels(self):
            msk = np.array([[True, False, True],
                            [False, False, False],
                            [True, False, True],
                            [True, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            mapper_data_to_2d = msk.grid_to_pixel()

            assert (mapper_data_to_2d[0] == np.array([0, 1])).all()
            assert (mapper_data_to_2d[1] == np.array([1, 0])).all()
            assert (mapper_data_to_2d[2] == np.array([1, 1])).all()
            assert (mapper_data_to_2d[3] == np.array([1, 2])).all()
            assert (mapper_data_to_2d[4] == np.array([2, 1])).all()
            assert (mapper_data_to_2d[5] == np.array([3, 2])).all()

    class TestMapperSparsePixels(object):

        # TODO : These tests are over crowded, should break up into more self contained things.

        def test__7x7_circle_mask__five_central_pixels__sparse_grid_size_1(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 1, 2, 3, 4])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 1, 2, 3, 4])).all()

        def test__7x7_circle_mask__sparse_grid_size_1(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.arange(21)).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.arange(21)).all()

        def test__7x7_rectangle_mask__sparse_grid_size_1(self):
            msk = np.array([[False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False],
                            [False, False, False, False, False, False, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.arange(49)).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.arange(49)).all()

        def test__7x7_circle_mask__sparse_grid_size_2(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([4, 6, 14, 16])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1,
                                                                           1, 2, 2, 2, 3, 3, 2, 2, 3])).all()

        def test__8x8_sporadic_mask__sparse_grid_size_2(self):
            msk = np.array([[True, True, True, True, True, True, False, False],
                            [True, True, False, False, False, True, False, False],
                            [True, False, False, False, False, False, False, False],
                            [True, False, False, False, False, False, False, False],
                            [True, False, False, False, False, False, False, False],
                            [True, True, False, False, False, True, False, False],
                            [True, True, True, True, True, True, False, False],
                            [True, True, False, False, False, True, False, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 8, 10, 12, 22, 24, 26, 33])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 0, 1, 1, 2, 0, 0, 1, 1, 1, 2, 2, 3, 3,
                                                                           1, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 4,
                                                                           4, 5, 6, 6,
                                                                           7, 7, 4, 4, 7, 7, 7])).all()

        def test__7x7_circle_mask_trues_on_even_values__sparse_grid_size_2(self):
            msk = np.array([[False, True, False, True, False, True, False],
                            [True, True, True, True, True, True, True],
                            [False, True, False, True, False, True, False],
                            [True, True, True, True, True, True, True],
                            [False, True, False, True, False, True, False],
                            [True, True, True, True, True, True, True],
                            [False, True, False, True, False, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.arange(16)).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.arange(16)).all()

        def test__7x7_circle_mask__sparse_grid_size_3(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 3)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([10])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array(
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])).all()

        def test__7x7_circle_mask_more_points_added__sparse_grid_size_3(self):
            msk = np.array([[False, True, True, False, True, False, False],
                            [True, True, False, False, False, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, False],
                            [True, False, False, False, False, False, True],
                            [True, True, False, False, False, True, True],
                            [True, True, True, True, True, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 3)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 1, 3, 14, 17, 26])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array(
                [0, 1, 2, 2, 1, 1, 1, 0, 3, 3, 3, 4, 3, 3, 3, 3, 4, 4, 3, 3, 3,
                 3, 4, 3, 3, 3, 5])).all()

        def test__7x7_mask_trues_on_values_which_divide_by_3__sparse_grid_size_3(self):
            msk = np.array([[False, True, True, False, True, True, False],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [False, True, True, False, True, True, False],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [False, True, True, False, True, True, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 3)

            assert (grid_cluster_pixelization.sparse_to_image == np.arange(9)).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.arange(9)).all()

        def test__8x8_mask_trues_on_values_which_divide_by_3_and_other_values__sparse_grid_size_3(self):
            msk = np.array([[False, True, False, False, True, True, False],
                            [True, True, True, True, True, True, True],
                            [True, True, False, False, False, True, True],
                            [False, True, True, False, True, True, False],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [False, False, False, False, False, False, False]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 3)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 2, 3, 7, 8, 9, 10, 13, 16])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array(
                [0, 1, 1, 2, 4, 4, 4, 3, 4, 5, 6, 6, 7, 7, 7, 8, 8])).all()

        def test__8x7__five_central_pixels__sparse_grid_size_1(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 1, 2, 3, 4])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 1, 2, 3, 4])).all()

        def test__8x7__five_central_pixels_2__sparse_grid_size_1(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 1, 2, 3, 4])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 1, 2, 3, 4])).all()

        def test__8x7__five_central_pixels__sparse_grid_size_2(self):
            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True],
                            [True, False, False, False, False, False, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([1, 3])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 0, 0, 1, 1, 0, 0, 0, 1, 1])).all()

        def test__7x8__five_central_pixels__sparse_grid_size_1(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 1)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([0, 1, 2, 3, 4])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 1, 2, 3, 4])).all()

        def test__7x8__five_central_pixels__sparse_grid_size_2(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([1, 3])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array([0, 0, 0, 1, 1, 0, 0, 0, 1, 1])).all()

        def test__7x8__more_central_pixels__sparse_grid_size_2(self):
            msk = np.array([[True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, False, False, False, False, False, True, True],
                            [True, True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            grid_cluster_pixelization = mask.SparseMask(msk, 2)

            assert (grid_cluster_pixelization.sparse_to_image == np.array([1, 3, 11, 13])).all()
            assert (grid_cluster_pixelization.image_to_sparse == np.array(
                [0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3])).all()


class TestSubCoordinateGrid(object):
    def test_sub_coordinate_grid(self, sub_coordinate_grid):
        assert sub_coordinate_grid.shape == (5, 2)
        assert (sub_coordinate_grid == np.array([[-1, 0], [0, -1], [0, 0], [0, 1], [1, 0]])).all()

    def test_sub_to_pixel(self, sub_coordinate_grid):
        assert (sub_coordinate_grid.sub_to_image == np.array(range(5))).all()

    def test_sub_data_to_image(self, sub_coordinate_grid):
        assert (sub_coordinate_grid.sub_data_to_image(np.array(range(5))) == np.array(range(5))).all()

    def test_setup_mappings_using_mask(self):
        msk = np.array([[True, False, True],
                        [False, False, False],
                        [True, False, True]])

        msk = mask.Mask(msk, pixel_scale=3.0)

        sub_coordinate_grid = mask.SubGrid.from_mask(msk, 2)

        assert sub_coordinate_grid.sub_grid_size == 2
        assert sub_coordinate_grid.sub_grid_fraction == (1.0 / 4.0)

        assert (sub_coordinate_grid.sub_to_image == np.array(
            [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])).all()


@pytest.fixture(name="memoizer")
def make_memoizer():
    return mask.Memoizer()


class TestMemoizer(object):
    def test_storing(self, memoizer):
        @memoizer
        def func(arg):
            return "result for {}".format(arg)

        func(1)
        func(2)
        func(1)

        assert memoizer.results == {"('arg', 1)": "result for 1", "('arg', 2)": "result for 2"}
        assert memoizer.calls == 2

    def test_multiple_arguments(self, memoizer):
        @memoizer
        def func(arg1, arg2):
            return arg1 * arg2

        func(1, 2)
        func(2, 1)
        func(1, 2)

        assert memoizer.results == {"('arg1', 1), ('arg2', 2)": 2, "('arg1', 2), ('arg2', 1)": 2}
        assert memoizer.calls == 2

    def test_key_word_arguments(self, memoizer):
        @memoizer
        def func(arg1=0, arg2=0):
            return arg1 * arg2

        func(arg1=1)
        func(arg2=1)
        func(arg1=1)
        func(arg1=1, arg2=1)

        assert memoizer.results == {"('arg1', 1)": 0, "('arg2', 1)": 0, "('arg1', 1), ('arg2', 1)": 1}
        assert memoizer.calls == 3

    def test_key_word_for_positional(self, memoizer):
        @memoizer
        def func(arg):
            return "result for {}".format(arg)

        func(1)
        func(arg=2)
        func(arg=1)

        assert memoizer.calls == 2

    def test_methods(self, memoizer):
        class Class(object):
            def __init__(self, value):
                self.value = value

            @memoizer
            def method(self):
                return self.value

        one = Class(1)
        two = Class(2)

        assert one.method() == 1
        assert two.method() == 2


@pytest.fixture(name="grids")
def make_grids(centre_mask):
    return mask.GridCollection.from_mask_sub_grid_size_and_blurring_shape(centre_mask, 2, (3, 3))


class TestGridCollection(object):
    def test_grids(self, grids):
        assert (grids.image == np.array([[0., 0.]])).all()
        np.testing.assert_almost_equal(grids.sub, np.array([[-0.16666667, -0.16666667],
                                                                                        [-0.16666667, 0.16666667],
                                                                                        [0.16666667, -0.16666667],
                                                                                        [0.16666667, 0.16666667]]))
        assert (grids.blurring == np.array([[-1., -1.],
                                                                   [-1., 0.],
                                                                   [-1., 1.],
                                                                   [0., -1.],
                                                                   [0., 1.],
                                                                   [1., -1.],
                                                                   [1., 0.],
                                                                   [1., 1.]])).all()

    def test_apply_function(self, grids):
        def add_one(coords):
            return np.add(1, coords)

        new_collection = grids.apply_function(add_one)
        assert isinstance(new_collection, mask.GridCollection)
        assert (new_collection.image == np.add(1, np.array([[0., 0.]]))).all()
        np.testing.assert_almost_equal(new_collection.sub, np.add(1, np.array([[-0.16666667, -0.16666667],
                                                                               [-0.16666667, 0.16666667],
                                                                               [0.16666667, -0.16666667],
                                                                               [0.16666667, 0.16666667]])))
        assert (new_collection.blurring == np.add(1, np.array([[-1., -1.],
                                                               [-1., 0.],
                                                               [-1., 1.],
                                                               [0., -1.],
                                                               [0., 1.],
                                                               [1., -1.],
                                                               [1., 0.],
                                                               [1., 1.]]))).all()

    def test_map_function(self, grids):
        def add_number(coords, number):
            return np.add(coords, number)

        new_collection = grids.map_function(add_number, [1, 2, 3])

        assert isinstance(new_collection, mask.GridCollection)
        assert (new_collection.image == np.add(1, np.array([[0., 0.]]))).all()
        np.testing.assert_almost_equal(new_collection.sub, np.add(2, np.array([[-0.16666667, -0.16666667],
                                                                               [-0.16666667, 0.16666667],
                                                                               [0.16666667, -0.16666667],
                                                                               [0.16666667, 0.16666667]])))
        assert (new_collection.blurring == np.add(3, np.array([[-1., -1.],
                                                               [-1., 0.],
                                                               [-1., 1.],
                                                               [0., -1.],
                                                               [0., 1.],
                                                               [1., -1.],
                                                               [1., 0.],
                                                               [1., 1.]]))).all()


class TestBorderCollection(object):

    class TestSetup:

        def test__simple_setup_using_constructor(self):

            image_border = mask.ImageGridBorder(arr=np.array([1, 2, 5]), polynomial_degree=4, centre=(1.0, 1.0))
            sub_border = mask.SubGridBorder(arr=np.array([1, 2, 3]), polynomial_degree=2, centre=(0.0, 1.0))

            border_collection = mask.BorderCollection(image=image_border, sub=sub_border)

            assert (border_collection.image == np.array([1, 2, 5])).all()
            assert border_collection.image.polynomial_degree == 4
            assert border_collection.image.centre == (1.0, 1.0)

            assert (border_collection.sub == np.array([1, 2, 3])).all()
            assert border_collection.sub.polynomial_degree == 2
            assert border_collection.sub.centre == (0.0, 1.0)

        def test__setup_from_mask(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border_collection = mask.BorderCollection.from_mask_and_sub_grid_size(mask=msk, sub_grid_size=2)

            assert (border_collection.image == np.array([0, 1])).all()
            assert (border_collection.sub == np.array([0, 5])).all()

    class TestRelocatedGridsFromGrids:
        
        def test__simple_case__new_grids_have_relocates(self):
            
            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            image_grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            image_grid = image_grid_circle
            image_grid.append(np.array([0.1, 0.0]))
            image_grid.append(np.array([-0.2, -0.3]))
            image_grid.append(np.array([0.5, 0.4]))
            image_grid.append(np.array([0.7, -0.1]))
            image_grid = np.asarray(image_grid)

            image_border = mask.ImageGridBorder(arr=np.arange(32), polynomial_degree=3)

            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            sub_grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            sub_grid = sub_grid_circle
            sub_grid.append(np.array([2.5, 0.0]))
            sub_grid.append(np.array([0.0, 3.0]))
            sub_grid.append(np.array([-2.5, 0.0]))
            sub_grid.append(np.array([-5.0, 5.0]))
            sub_grid = np.asarray(sub_grid)

            sub_border = mask.SubGridBorder(arr=np.arange(32), polynomial_degree=3)

            borders = mask.BorderCollection(image=image_border, sub=sub_border)

            grids = mask.GridCollection(image=image_grid, sub=sub_grid, blurring=None)

            relocated_grids = borders.relocated_grids_from_grids(grids)

            assert relocated_grids.image[0:32] == pytest.approx(np.asarray(image_grid_circle)[0:32], 1e-3)
            assert relocated_grids.image[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.image[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.image[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.image[35] == pytest.approx(np.array([0.7, -0.1]), 1e-3)

            assert relocated_grids.sub[0:32] == pytest.approx(np.asarray(sub_grid_circle)[0:32], 1e-3)
            assert relocated_grids.sub[32] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert relocated_grids.sub[33] == pytest.approx(np.array([0.0, 1.0]), 1e-3)
            assert relocated_grids.sub[34] == pytest.approx(np.array([-1.0, 0.0]), 1e-3)
            assert relocated_grids.sub[35] == pytest.approx(np.array([-0.707, 0.707]), 1e-3)


class TestImageGridBorder(object):

    class TestFromMask:

        def test__simple_mask_border_pixel_is_pixel(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border = mask.ImageGridBorder.from_mask(msk)

            assert (border == np.array([0, 1])).all()

    class TestThetasAndRadii:

        def test__four_grid_in_circle__all_in_border__correct_radii_and_thetas(self):

            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0, -1.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3)
            radii = border.grid_to_radii(grid)
            thetas = border.grid_to_thetas(grid)

            assert (radii == np.array([1.0, 1.0, 1.0, 1.0])).all()
            assert (thetas == np.array([0.0, 90.0, 180.0, 270.0])).all()

        def test__other_thetas_radii(self):
            grid = np.array([[2.0, 0.0], [2.0, 2.0], [-1.0, -1.0], [0.0, -3.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3)
            radii = border.grid_to_radii(grid)
            thetas = border.grid_to_thetas(grid)

            assert (radii == np.array([2.0, 2.0 * np.sqrt(2), np.sqrt(2.0), 3.0])).all()
            assert (thetas == np.array([0.0, 45.0, 225.0, 270.0])).all()

        def test__border_centre_offset__grid_same_r_and_theta_shifted(self):

            grid = np.array([[2.0, 1.0], [1.0, 2.0], [0.0, 1.0], [1.0, 0.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3, centre=(1.0, 1.0))
            radii = border.grid_to_radii(grid)
            thetas = border.grid_to_thetas(grid)

            assert (radii == np.array([1.0, 1.0, 1.0, 1.0])).all()
            assert (thetas == np.array([0.0, 90.0, 180.0, 270.0])).all()

    class TestBorderPolynomialFit(object):

        def test__four_grid_in_circle__thetas_at_radius_are_each_grid_radius(self):

            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0, -1.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3)
            poly = border.polynomial_fit_to_border(grid)

            assert np.polyval(poly, 0.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 90.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 180.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 270.0) == pytest.approx(1.0, 1e-3)

        def test__eight_grid_in_circle__thetas_at_each_grid_are_the_radius(self):

            grid = np.array([[1.0, 0.0], [0.5 * np.sqrt(2), 0.5 * np.sqrt(2)],
                                    [0.0, 1.0], [-0.5 * np.sqrt(2), 0.5 * np.sqrt(2)],
                                    [-1.0, 0.0], [-0.5 * np.sqrt(2), -0.5 * np.sqrt(2)],
                                    [0.0, -1.0], [0.5 * np.sqrt(2), -0.5 * np.sqrt(2)]])

            border = mask.ImageGridBorder(arr=
                                      np.arange(8), polynomial_degree=3)
            poly = border.polynomial_fit_to_border(grid)

            assert np.polyval(poly, 0.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 45.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 90.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 135.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 180.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 225.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 270.0) == pytest.approx(1.0, 1e-3)
            assert np.polyval(poly, 315.0) == pytest.approx(1.0, 1e-3)

    class TestMoveFactors(object):

        def test__inside_border__move_factor_is_1(self):
            
            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0, -1.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3)
            move_factors = border.move_factors_from_grid(grid)

            assert move_factors[0] == pytest.approx(1.0, 1e-4)
            assert move_factors[1] == pytest.approx(1.0, 1e-4)
            assert move_factors[2] == pytest.approx(1.0, 1e-4)
            assert move_factors[3] == pytest.approx(1.0, 1e-4)

        def test__outside_border_double_its_radius__move_factor_is_05(self):

            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0, -1.0],
                             [2.0, 0.0], [0.0, 2.0], [-2.0, 0.0], [0.0, -2.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3)
            move_factors = border.move_factors_from_grid(grid)

            assert move_factors[0] == pytest.approx(1.0, 1e-4)
            assert move_factors[1] == pytest.approx(1.0, 1e-4)
            assert move_factors[2] == pytest.approx(1.0, 1e-4)
            assert move_factors[3] == pytest.approx(1.0, 1e-4)
            assert move_factors[4] == pytest.approx(0.5, 1e-4)
            assert move_factors[5] == pytest.approx(0.5, 1e-4)
            assert move_factors[6] == pytest.approx(0.5, 1e-4)
            assert move_factors[7] == pytest.approx(0.5, 1e-4)

        def test__outside_border_as_above__but_shift_for_source_plane_centre(self):

            grid = np.array([[2.0, 1.0], [1.0, 2.0], [0.0, 1.0], [1.0, 0.0],
                             [3.0, 1.0], [1.0, 3.0], [1.0, 3.0], [3.0, 1.0]])

            border = mask.ImageGridBorder(arr=np.arange(4), polynomial_degree=3, centre=(1.0, 1.0))
            move_factors = border.move_factors_from_grid(grid)

            assert move_factors[0] == pytest.approx(1.0, 1e-4)
            assert move_factors[1] == pytest.approx(1.0, 1e-4)
            assert move_factors[2] == pytest.approx(1.0, 1e-4)
            assert move_factors[3] == pytest.approx(1.0, 1e-4)
            assert move_factors[4] == pytest.approx(0.5, 1e-4)
            assert move_factors[5] == pytest.approx(0.5, 1e-4)
            assert move_factors[6] == pytest.approx(0.5, 1e-4)
            assert move_factors[7] == pytest.approx(0.5, 1e-4)

    class TestRelocateCoordinates(object):

        def test__inside_border_no_relocations(self):

            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            grid = grid_circle
            grid.append(np.array([0.1, 0.0]))
            grid.append(np.array([-0.2, -0.3]))
            grid.append(np.array([0.5, 0.4]))
            grid.append(np.array([0.7, -0.1]))
            grid = np.asarray(grid)

            border = mask.ImageGridBorder(arr=np.arange(32), polynomial_degree=3)
            relocated_grid = border.relocated_grid_from_grid(grid)

            assert relocated_grid[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grid[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grid[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grid[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grid[35] == pytest.approx(np.array([0.7, -0.1]), 1e-3)

        def test__outside_border_simple_cases__relocates_to_source_border(self):

            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            grid = grid_circle
            grid.append(np.array([2.5, 0.0]))
            grid.append(np.array([0.0, 3.0]))
            grid.append(np.array([-2.5, 0.0]))
            grid.append(np.array([-5.0, 5.0]))
            grid = np.asarray(grid)

            border = mask.ImageGridBorder(arr=np.arange(32), polynomial_degree=3)
            relocated_grid = border.relocated_grid_from_grid(grid)

            assert relocated_grid[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grid[32] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert relocated_grid[33] == pytest.approx(np.array([0.0, 1.0]), 1e-3)
            assert relocated_grid[34] == pytest.approx(np.array([-1.0, 0.0]), 1e-3)
            assert relocated_grid[35] == pytest.approx(np.array([-0.707, 0.707]), 1e-3)

        def test__6_grid_total__2_outside_border__different_border__relocate_to_source_border(self):

            grid = np.array([[1.0, 0.0], [20., 20.], [0.0, 1.0], [-1.0, 0.0], [0.0, -1.0], [1.0, 1.0]])
            border_pixels = np.array([0, 2, 3, 4])

            border = mask.ImageGridBorder(border_pixels, polynomial_degree=3)

            relocated_grid = border.relocated_grid_from_grid(grid)

            assert relocated_grid[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grid[1] == pytest.approx(np.array([0.7071, 0.7071]), 1e-3)
            assert relocated_grid[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grid[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grid[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grid[5] == pytest.approx(np.array([0.7071, 0.7071]), 1e-3)


class TestSubGridBorder(object):

    class TestFromMask:

        def test__simple_mask_border_pixel_is_pixel(self):

            msk = np.array([[True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, False, False, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True],
                            [True, True, True, True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border = mask.SubGridBorder.from_mask(msk, sub_grid_size=2)

            assert (border == np.array([0, 5])).all()






from autolens.pixelization import reconstruction

import numpy as np
import pytest


class TestCovarianceMatrixFromBlurred(object):

    def test__simple_blurred_mapping_matrix__correct_covariance_matrix(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        noise_vector = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        cov = recon.covariance_matrix_from_blurred_mapping_jit(blurred_mapping_matrix, noise_vector)

        assert (cov == np.array([[2.0, 1.0, 0.0],
                                 [1.0, 3.0, 1.0],
                                 [0.0, 1.0, 1.0]])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values__correct_covariance_matrix(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        noise_vector = np.array([2.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        cov = recon.covariance_matrix_from_blurred_mapping_jit(blurred_mapping_matrix, noise_vector)

        assert (cov == np.array([[1.25, 0.25, 0.0],
                                 [0.25, 2.25, 1.0],
                                 [0.0, 1.0, 1.0]])).all()

    def test__jitted_and_normal(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [2.0, 6.0, 1.0],
                                           [0.0, 1.0, 8.0],
                                           [3.0, 1.0, 1.0],
                                           [4.0, 3.0, 0.0],
                                           [0.0, 0.0, 3.0]])

        noise_vector = np.array([3.0, 1.0, 4.0, 1.0, 10.0, 1.0])

        cov = recon.covariance_matrix_from_blurred_mapping_jit(blurred_mapping_matrix, noise_vector)
        cov_jitted = recon.covariance_matrix_from_blurred_mapping_jit(blurred_mapping_matrix, noise_vector)

        assert (cov == cov_jitted).all()


class TestDataVectorFromData(object):

    def test__simple_blurred_mapping_matrix__correct_d_matrix(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image_vector = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        noise_vector = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        d = recon.data_vector_from_blurred_mapping_and_data_jit(blurred_mapping_matrix, image_vector, noise_vector)

        assert (d == np.array([2.0, 3.0, 1.0])).all()

    def test__simple_blurred_mapping_matrix__change_image_values__correct_d_matrix(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image_vector = np.array([3.0, 1.0, 1.0, 10.0, 1.0, 1.0])
        noise_vector = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        d = recon.data_vector_from_blurred_mapping_and_data_jit(blurred_mapping_matrix, image_vector, noise_vector)

        assert (d == np.array([4.0, 14.0, 10.0])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values__correct_d_matrix(self):

        recon = reconstruction.Reconstructor(mapping=np.ones((6,3)), regularization=np.ones((1,1)),
                                             image_to_pix=np.ones((1,1)), sub_to_pix=np.ones((1,1)))

        blurred_mapping_matrix = np.array([[1.0, 1.0, 0.0],
                                           [1.0, 0.0, 0.0],
                                           [0.0, 1.0, 0.0],
                                           [0.0, 1.0, 1.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])

        image_vector = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_vector = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        d = recon.data_vector_from_blurred_mapping_and_data_jit(blurred_mapping_matrix, image_vector, noise_vector)

        assert (d == np.array([2.0, 3.0, 1.0])).all()


class TestComputeRegularizationTerm:

    def test__solution_all_1s__regularization_matrix_simple(self):
        solution = np.array([1.0, 1.0, 1.0])

        regularization_matrix = np.array([[1.0, 0.0, 0.0],
                                          [0.0, 1.0, 0.0],
                                          [0.0, 0.0, 1.0]])

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [1.0, 1.0, 1.0] * [1.0, 1.0, 1.0] = [(1.0*1.0) + (1.0*0.0) + (1.0*0.0)] = [1.0, 1.0, 1.0]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*1.0) + (1.0*0.0)]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*0.0) + (1.0*1.0)]

        # (s_T * H) * s = [1.0, 1.0, 1.0] * [1.0] = 3.0
        #                                   [1.0]
        #                                   [1.0]

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=None,
                                                regularization=regularization_matrix,
                                                covariance=None, covariance_regularization=None,
                                                reconstruction=solution)

        assert pix_fit.regularization_term_from_reconstruction() == 3.0

    def test__solution_and_regularization_matrix_range_of_values(self):
        solution = np.array([2.0, 3.0, 5.0])

        regularization_matrix = np.array([[2.0, -1.0, 0.0],
                                          [-1.0, 2.0, -1.0],
                                          [0.0, -1.0, 2.0]])

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [2.0, 3.0, 5.0] * [2.0,  -1.0,  0.0] = [(2.0* 2.0) + (3.0*-1.0) + (5.0 *0.0)] = [1.0, -1.0, 7.0]
        #                             [-1.0,  2.0, -1.0]   [(2.0*-1.0) + (3.0* 2.0) + (5.0*-1.0)]
        #                             [ 0.0, -1.0,  2.0]   [(2.0* 0.0) + (3.0*-1.0) + (5.0 *2.0)]

        # (s_T * H) * s = [1.0, -1.0, 7.0] * [2.0] = 34.0
        #                                    [3.0]
        #                                    [5.0]

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=None,
                                                regularization=regularization_matrix,
                                                covariance=None, covariance_regularization=None,
                                                reconstruction=solution)

        assert pix_fit.regularization_term_from_reconstruction() == 34.0


class TestLogDetMatrix:

    def test__determinant_of_positive_definite_matrix_via_cholesky(self):
        matrix = np.array([[1.0, 0.0, 0.0],
                           [0.0, 1.0, 0.0],
                           [0.0, 0.0, 1.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=None, regularization=None,
                                                covariance=None, covariance_regularization=None, reconstruction=None)

        assert log_determinant == pytest.approx(pix_fit.log_determinant_of_matrix_cholesky(matrix), 1e-4)

    def test__determinant_of_positive_definite_matrix_2_via_cholesky(self):
        matrix = np.array([[2.0, -1.0, 0.0],
                           [-1.0, 2.0, -1.0],
                           [0.0, -1.0, 2.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=None, regularization=None,
                                                covariance=None, covariance_regularization=None, reconstruction=None)

        assert log_determinant == pytest.approx(pix_fit.log_determinant_of_matrix_cholesky(matrix), 1e-4)


class TestModelImageFromSolution:

    def test__solution_all_1s__simple_blurred_mapping__correct_model_image(self):
        solution = np.array([1.0, 1.0, 1.0, 1.0])

        blurred_mapping = np.array([[1.0, 1.0, 1.0, 1.0],
                                    [1.0, 0.0, 1.0, 1.0],
                                    [1.0, 0.0, 0.0, 0.0]])

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=blurred_mapping,
                                                regularization=None,
                                                covariance=None, covariance_regularization=None,
                                                reconstruction=solution)

        model_image = pix_fit.model_image_from_reconstruction_jit()

        # Image pixel 0 maps to 4 pixs pixxels -> value is 4.0
        # Image pixel 1 maps to 3 pixs pixxels -> value is 3.0
        # Image pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (model_image == np.array([4.0, 3.0, 1.0])).all()

    def test__solution_different_values__simple_blurred_mapping__correct_model_image(self):
        solution = np.array([1.0, 2.0, 3.0, 4.0])

        blurred_mapping = np.array([[1.0, 1.0, 1.0, 1.0],
                                    [1.0, 0.0, 1.0, 1.0],
                                    [1.0, 0.0, 0.0, 0.0]])

        pix_fit = reconstruction.Reconstruction(data_vector=None, blurred_mapping=blurred_mapping,
                                                regularization=None,
                                                covariance=None, covariance_regularization=None,
                                                reconstruction=solution)

        model_image = pix_fit.model_image_from_reconstruction_jit()

        # Image pixel 0 maps to 4 pixs pixxels -> value is 1.0 + 2.0 + 3.0 + 4.0 = 10.0
        # Image pixel 1 maps to 3 pixs pixxels -> value is 1.0 + 3.0 + 4.0
        # Image pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (model_image == np.array([10.0, 8.0, 1.0])).all()



import sys
sys.path.append("../")
import os
import numpy as np
import matplotlib.pyplot as plt

from autolens.imaging import image
from autolens.imaging import mask
from autolens.profiles import light_profiles as lp
from autolens.profiles import mass_profiles as mp
from autolens.analysis import ray_tracing
from autolens.analysis import galaxy
from imaging import array_util

path =  "{}/".format(os.path.dirname(os.path.realpath(__file__)))
output_path = "{}/../data/integration/".format(os.path.dirname(os.path.realpath(__file__)))

lens_name = 'hst_1/'
psf_size = (21, 21)
pixel_scale = 0.05

psf = image.PSF.from_fits(file_path=path+'../profiling/data/psf', hdu=3, pixel_scale=pixel_scale).trim(psf_size)
psf = psf.renormalize()
ma = mask.Mask.for_simulate(shape_arc_seconds=(15.0, 15.0), pixel_scale=pixel_scale, psf_size=psf_size)

image_plane_grids = mask.GridCollection.from_mask_sub_grid_size_and_blurring_shape(mask=ma, sub_grid_size=1,
                                                                                   blurring_shape=psf_size)

### Setup the ray tracing model, and use to generate the 2D galaxy image_coords ###

# sersic_lens = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.5, effective_radius=1.6,
#                              sersic_index=4.0)
# isothermal = mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=1.3)
# sersic_source = lp.EllipticalSersic(centre=(0.3, 0.2), axis_ratio=0.6, phi=45.0, intensity=2.0, effective_radius=1.0,
#                              sersic_index=1.0)

sersic_lens = lp.EllipticalSersic(centre=(0.1, -0.1), axis_ratio=0.7, phi=90.0, intensity=1.0, effective_radius=1.3,
                             sersic_index=3.0)
isothermal = mp.EllipticalIsothermal(centre=(0.1, -0.1), axis_ratio=0.6, phi=30.0, einstein_radius=0.9)
sersic_source = lp.EllipticalSersic(centre=(-0.2, 0.2), axis_ratio=0.9, phi=45.0, intensity=1.0, effective_radius=1.0,
                             sersic_index=2.0)

lens_galaxy = galaxy.Galaxy(light_profile=sersic_lens, mass_profile=isothermal)
source_galaxy = galaxy.Galaxy(light_profile=sersic_source)

ray_trace = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                               image_plane_grids=image_plane_grids)

galaxy_image_1d = ray_trace.generate_image_of_galaxy_light_profiles()
galaxy_image_2d = ma.map_to_2d(galaxy_image_1d)

### Setup as a simulated image_coords and output as a fits for an analysis ###

shape = galaxy_image_2d.shape
sim_image = image.Image.simulate(array=galaxy_image_2d, effective_exposure_time=200.0*np.ones(shape),
                                 pixel_scale=pixel_scale,
                                 background_sky_map=10.0*np.ones(shape), psf=psf, include_poisson_noise=True, seed=1)

print(np.max(sim_image))
print(np.max(sim_image) / np.max(sim_image.estimated_noise))

plt.imshow(sim_image)
plt.show()

if os.path.exists(output_path + lens_name) == False:
    os.makedirs(output_path + lens_name)

array_util.numpy_array_to_fits(sim_image, file_path=output_path + lens_name + 'image')
array_util.numpy_array_to_fits(sim_image.estimated_noise, file_path=output_path + lens_name + 'noise')
array_util.numpy_array_to_fits(sim_image.effective_exposure_time, file_path=output_path + lens_name + 'exposure_time')
array_util.numpy_array_to_fits(psf, file_path=output_path + lens_name + '/psf')


from autolens.analysis import galaxy_prior as gp
from autolens.analysis import galaxy as g
from autolens.analysis import ray_tracing
from autolens.imaging import mask as msk
from autolens.imaging import masked_image as mi
from autolens.imaging import image as img
from autolens.analysis import fitting
from autolens.autopipe import non_linear
from autolens.autopipe import model_mapper as mm
import numpy as np
from autolens.pixelization import pixelization as px
import inspect
import logging

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


def default_mask_function(image):
    return msk.Mask.circular(image.shape_arc_seconds, image.pixel_scale, 3)


class ResultsCollection(list):
    def __init__(self, results):
        super().__init__(results)

    @property
    def last(self):
        if len(self) > 0:
            return self[-1]
        return None

    @property
    def first(self):
        if len(self) > 0:
            return self[0]
        return None


class Phase(object):
    def __init__(self, optimizer_class=non_linear.DownhillSimplex, sub_grid_size=1,
                 mask_function=default_mask_function, name=None):
        """
        A phase in an analysis pipeline. Uses the set non_linear optimizer to try to fit models and images passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """
        self.optimizer = optimizer_class(name=name)
        self.sub_grid_size = sub_grid_size
        self.mask_function = mask_function
        self.name = name

    def run(self, image, last_results=None):
        """
        Run this phase.

        Parameters
        ----------
        last_results: ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        image: img.Image
            An image that has been masked

        Returns
        -------
        result: non_linear.Result
            A result object comprising the best fit model and other data.
        """
        analysis = self.make_analysis(image=image, previous_results=last_results)
        result = self.optimizer.fit(analysis)
        return self.__class__.Result(result.constant, result.likelihood, result.variable,
                                     galaxy_images=analysis.galaxy_images_for_model(result.constant))

    def make_analysis(self, image, previous_results=None):
        """
        Create an analysis object. Also calls the prior passing and image modifying functions to allow child classes to
        change the behaviour of the phase.

        Parameters
        ----------
        image: im.Image
            An image that has been masked
        previous_results: ResultsCollection
            The result from the previous phase

        Returns
        -------
        analysis: Analysis
            An analysis object that the non-linear optimizer calls to determine the fit of a set of values
        """
        mask = self.mask_function(image)
        image = self.modify_image(image, previous_results)
        masked_image = mi.MaskedImage(image, mask, sub_grid_size=self.sub_grid_size)
        self.pass_priors(previous_results)

        analysis = self.__class__.Analysis(masked_image=masked_image, previous_results=previous_results)
        return analysis

    @property
    def constant(self):
        """
        Convenience method

        Returns
        -------
        ModelInstance
            A model instance comprising all the constant objects in this analysis
        """
        return self.optimizer.constant

    @property
    def variable(self):
        """
        Convenience method

        Returns
        -------
        ModelMapper
            A model mapper comprising all the variable (prior) objects in this analysis
        """
        return self.optimizer.variable

    class Result(non_linear.Result):

        def __init__(self, constant, likelihood, variable, galaxy_images):
            """
            The result of a phase

            Parameters
            ----------

            galaxy_images: [ndarray]
                A collection of images created by each individual galaxy which taken together form the full model image
            """
            super(Phase.Result, self).__init__(constant, likelihood, variable)
            self.galaxy_images = galaxy_images

        @property
        def model_image(self):
            return np.sum(np.stack((image for image in self.galaxy_images if image is not None)), axis=0)

    class Analysis(object):
        def __init__(self, previous_results,
                     masked_image):
            """
            An analysis object

            Parameters
            ----------
            previous_results: ResultsCollection
                The results of all previous phases
            masked_image: mi.MaskedImage
                An image that has been masked
            """
            self.previous_results = previous_results
            self.masked_image = masked_image

        @property
        def last_results(self):
            if self.previous_results is not None:
                return self.previous_results.last

        def fit(self, **kwargs):
            """
            Determine the fitness of a particular model

            Parameters
            ----------
            kwargs: dict
                Dictionary of objects describing the model

            Returns
            -------
            fit: fitting.Fit
                How fit the model is and the model
            """
            raise NotImplementedError()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def modify_image(self, image, previous_results):
        """
        Customize an image. e.g. removing lens light.

        Parameters
        ----------
        image: img.Image
            An image that has been masked
        previous_results: ResultsCollection
            The result of the previous analysis

        Returns
        -------
        image: img.Image
            The modified image (not changed by default)
        """
        return image

    def pass_priors(self, previous_results):
        """
        Perform any prior or constant passing. This could involve setting model attributes equal to priors or constants
        from a previous phase.

        Parameters
        ----------
        previous_results: ResultsCollection
            The result of the previous phase
        """
        pass


def phase_property(name):
    """
    Create a property that is tied to the non_linear instance determines whether to set itself as a constant or
    variable.

    Parameters
    ----------
    name: str
        The name of this variable

    Returns
    -------
    property: property
        A property that appears to be an attribute of the phase but is really an attribute of constant or variable.
    """

    def fget(self):
        if hasattr(self.optimizer.constant, name):
            return getattr(self.optimizer.constant, name)
        elif hasattr(self.optimizer.variable, name):
            return getattr(self.optimizer.variable, name)

    def fset(self, value):
        if inspect.isclass(value) or isinstance(value, gp.GalaxyPrior):
            setattr(self.optimizer.variable, name, value)
            try:
                delattr(self.optimizer.constant, name)
            except AttributeError:
                pass
        else:
            setattr(self.optimizer.constant, name, value)
            try:
                delattr(self.optimizer.variable, name)
            except AttributeError:
                pass

    return property(fget=fget, fset=fset, doc=name)


class SourceLensPhase(Phase):
    lens_galaxy = phase_property("lens_galaxy")
    source_galaxy = phase_property("source_galaxy")

    def __init__(self,
                 lens_galaxy=None,
                 source_galaxy=None,
                 optimizer_class=non_linear.DownhillSimplex,
                 sub_grid_size=1,
                 mask_function=default_mask_function,
                 name="source_lens_phase"):
        """
        A phase with a simple source/lens model

        Parameters
        ----------
        lens_galaxy: g.Galaxy | gp.GalaxyPrior
            A galaxy that acts as a gravitational lens
        source_galaxy: g.Galaxy | gp.GalaxyPrior
            A galaxy that is being lensed
        optimizer_class: class
            The class of a non-linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """
        super().__init__(
            optimizer_class=optimizer_class,
            sub_grid_size=sub_grid_size,
            mask_function=mask_function,
            name=name)
        self.lens_galaxy = lens_galaxy
        self.source_galaxy = source_galaxy

    def modify_image(self, masked_image, last_result):
        """

        Parameters
        ----------
        masked_image: mi.MaskedImage
        last_result: SourceLensPhase.Result

        Returns
        -------

        """
        return masked_image

    class Result(Phase.Result):
        @property
        def lens_galaxy_image(self):
            return self.galaxy_images[0]

        @property
        def source_galaxy_image(self):
            return self.galaxy_images[1]

    class Analysis(Phase.Analysis):
        def __init__(self, masked_image, previous_results):
            super(SourceLensPhase.Analysis, self).__init__(previous_results, masked_image)
            self.model_image = None
            self.galaxy_images = None
            if self.last_results is not None:
                self.model_image = self.masked_image.mask.map_to_1d(previous_results.last.model_image)
                self.galaxy_images = list(map(self.masked_image.mask.map_to_1d, previous_results.last.galaxy_images))

        def fit(self, lens_galaxy=None, source_galaxy=None):
            """
            Determine the fit of a lens galaxy and source galaxy to the image in this analysis.

            Parameters
            ----------
            lens_galaxy: g.Galaxy
                The galaxy that acts as a gravitational lens
            source_galaxy: g.Galaxy
                The galaxy that produces the light that is being lensed

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model image itself
            """
            logger.debug(
                "\nRunning lens/source analysis for... \n\nLens Galaxy:\n{}\n\nSource Galaxy:\n{}\n\n".format(
                    lens_galaxy,
                    source_galaxy))

            tracer = ray_tracing.Tracer(
                [] if lens_galaxy is None else [lens_galaxy],
                [] if source_galaxy is None else [source_galaxy], self.masked_image.grids)
            fitter = fitting.Fitter(self.masked_image, tracer)

            if self.last_results is not None and tracer.all_with_hyper_galaxies:
                return fitter.fit_data_with_profiles_and_model_images(self.model_image,
                                                                      self.galaxy_images)

            return fitter.fit_data_with_profiles()

        def galaxy_images_for_model(self, model):
            """
            Generate images of galaxies for a set model.

            Parameters
            ----------
            model: mm.ModelInstance
                A model instance

            Returns
            -------
            galaxy_images: [ndarray]
                A list of images of galaxy components
            """

            def model_image(plane):
                if len(plane.galaxies) == 0:
                    return None
                return self.masked_image.map_to_2d(plane.galaxy_images[0])

            lens_galaxies = [] if model.lens_galaxy is None else [model.lens_galaxy]
            source_galaxies = [] if model.source_galaxy is None else [model.source_galaxy]
            tracer = ray_tracing.Tracer(lens_galaxies, source_galaxies, self.masked_image.grids)
            return model_image(tracer.image_plane), model_image(tracer.source_plane)


class PixelizedSourceLensPhase(SourceLensPhase):
    def __init__(self,
                 lens_galaxy=None,
                 pixelization=px.RectangularRegConst,
                 optimizer_class=non_linear.DownhillSimplex,
                 sub_grid_size=1,
                 mask_function=default_mask_function):
        super().__init__(lens_galaxy=lens_galaxy, source_galaxy=gp.GalaxyPrior(pixelization=pixelization),
                         optimizer_class=optimizer_class, sub_grid_size=sub_grid_size, mask_function=mask_function)

    class Analysis(SourceLensPhase.Analysis):

        def fit(self, lens_galaxy=None, source_galaxy=None):
            """
            Determine the fit of a lens galaxy and source galaxy to the image in this analysis.

            Parameters
            ----------
            lens_galaxy: g.Galaxy
                The galaxy that acts as a gravitational lens
            source_galaxy: g.Galaxy
                The galaxy that produces the light that is being lensed

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model image itself
            """
            logger.debug(
                "\nRunning lens/source analysis for... \n\nLens Galaxy:\n{}\n\nSource Galaxy:\n{}\n\n".format(
                    lens_galaxy,
                    source_galaxy))

            tracer = ray_tracing.Tracer([lens_galaxy], [source_galaxy], self.masked_image.grids)
            fitter = fitting.PixelizedFitter(self.masked_image, "TODO", tracer)

            if self.last_results is not None and tracer.all_with_hyper_galaxies:
                return fitter.fit_data_with_pixelization_profiles_and_model_images(self.last_results.model_image,
                                                                                   self.last_results.galaxy_images)

            return fitter.fit_data_with_pixelization_and_profiles()


class LensOnlyPhase(SourceLensPhase):
    def __init__(self,
                 lens_galaxy=None,
                 optimizer_class=non_linear.DownhillSimplex,
                 sub_grid_size=1,
                 mask_function=default_mask_function,
                 name="lens_only_phase"
                 ):
        super(LensOnlyPhase, self).__init__(lens_galaxy=lens_galaxy,
                                            optimizer_class=optimizer_class,
                                            sub_grid_size=sub_grid_size,
                                            mask_function=mask_function,
                                            name=name)


class SourceOnlyPhase(SourceLensPhase):
    def __init__(self,
                 source_galaxy=None,
                 optimizer_class=non_linear.DownhillSimplex,
                 sub_grid_size=1,
                 mask_function=default_mask_function
                 ):
        super(SourceOnlyPhase, self).__init__(source_galaxy=source_galaxy,
                                              optimizer_class=optimizer_class,
                                              sub_grid_size=sub_grid_size,
                                              mask_function=mask_function)


class SourceLensHyperGalaxyPhase(SourceLensPhase):
    # TODO: Perform hyper galaxy analyses for each hyper galaxy independently
    def pass_priors(self, previous_results):
        self.lens_galaxy = gp.GalaxyPrior.from_galaxy(
            previous_results.last.constant.lens_galaxy,
            hyper_galaxy=g.HyperGalaxy)
        self.source_galaxy = gp.GalaxyPrior.from_galaxy(
            previous_results.last.constant.source_galaxy,
            hyper_galaxy=g.HyperGalaxy)



import sys
sys.path.append("../")
import os
import numpy as np
import matplotlib.pyplot as plt

from autolens.imaging import image
from autolens.imaging import mask
from autolens.profiles import light_profiles as lp
from autolens.profiles import mass_profiles as mp
from autolens.analysis import ray_tracing
from autolens.analysis import galaxy
from imaging import array_util

path =  "{}/".format(os.path.dirname(os.path.realpath(__file__)))
output_path = "{}/../profiling/data/".format(os.path.dirname(os.path.realpath(__file__)))

psf_size = (51, 51)

lens_name = 'LSST/'
pixel_scale = 0.2

lens_name = 'Euclid/'
pixel_scale = 0.1

lens_name = 'HST/'
pixel_scale = 0.05

lens_name = 'HSTup/'
pixel_scale = 0.03

lens_name = 'AO/'
pixel_scale = 0.01

psf = image.PSF.from_fits(file_path=path+'../profiling/data/psf', hdu=3, pixel_scale=pixel_scale)
psf = psf.trim(psf_size)
ma = mask.Mask.for_simulate(shape_arc_seconds=(15.0, 15.0), pixel_scale=pixel_scale, psf_size=psf_size)

image_plane_grids = mask.GridCollection.from_mask_sub_grid_size_and_blurring_shape(mask=ma, sub_grid_size=4,
                                                                                   blurring_shape=psf_size)

### Setup the ray tracing model, and use to generate the 2D galaxy image_coords ###

sersic = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.5, effective_radius=1.3,
                             sersic_index=3.0)
isothermal = mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=1.4)

lens_galaxy = galaxy.Galaxy(light_profile=sersic, mass_profile=isothermal)
source_galaxy = galaxy.Galaxy(light_profile=sersic)

ray_trace = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                               image_plane_grids=image_plane_grids)

galaxy_image_1d = ray_trace.generate_image_of_galaxy_light_profiles()
galaxy_image_2d = ma.map_to_2d(galaxy_image_1d)

plt.imshow(galaxy_image_2d)
plt.show()

### Setup as a simulated image_coords and output as a fits for an analysis ###

shape = galaxy_image_2d.shape
sim_image = image.Image.simulate(array=galaxy_image_2d, effective_exposure_time=np.ones(shape), pixel_scale=pixel_scale,
                                 background_sky_map=np.ones(shape), psf=psf, include_poisson_noise=True, seed=1)

if os.path.exists(output_path + lens_name) == False:
    os.makedirs(output_path + lens_name)

array_util.numpy_array_to_fits(sim_image, file_path=output_path + lens_name + 'image')
array_util.numpy_array_to_fits(np.ones(sim_image.shape), file_path=output_path + lens_name + 'noise')
array_util.numpy_array_to_fits(np.ones(sim_image.shape), file_path=output_path + lens_name + 'exposure_time')
array_util.numpy_array_to_fits(psf, file_path=output_path + lens_name + '/psf')


import sys

sys.path.append("../../")
import numpy as np
from autolens.imaging import mask
from imaging import convolution
import time
import os

path = os.path.dirname(os.path.realpath(__file__))

def load(name):
    return np.load("{}/{}.npy".format(path, name))

grid = load("deflection_data/grid")

psf_shape = (11, 11)

ma = mask.Mask.for_simulate(shape_arc_seconds=(4.0, 4.0), pixel_scale=0.1, psf_size=psf_shape)

data = ma.map_to_1d(np.ones(ma.shape))

mapping = np.ones((len(data), 60))

frame = convolution.Convolver(mask=ma)
convolver = frame.convolver_for_kernel_shape(kernel_shape=psf_shape)
# This PSF leads to no blurring, so equivalent to being off.
kernel_convolver = convolver.convolver_for_kernel(kernel=np.ones(psf_shape))

kernel_convolver.convolve_mapping_matrix_jit(mapping)
repeats = 1

def tick_toc(func):
    def wrapper():
        start = time.time()
        for _ in range(repeats):
            func()

        diff = time.time() - start
        print("{}: {}".format(func.__name__, diff))

    return wrapper

@tick_toc
def current_solution():

    kernel_convolver.convolve_mapping_matrix_jit(mapping)


@tick_toc
def jitted_solution():
    kernel_convolver.convolve_mapping_matrix_jit(mapping)

if __name__ == "__main__":
    current_solution()
    jitted_solution()


import numpy as np

# TODO : Think carefully about demagnified centra pixels.


class InterpolationGeometry(object):

    def __init__(self, y_min, y_max, x_min, x_max, y_pixel_scale, x_pixel_scale):
        """The geometry of a rectangular grid, defining where the grids top-left, top-right, bottom-left and \
        bottom-right corners are in arc seconds. The arc-second size of each rectangular pixel is also computed.

        Parameters
        -----------

        """
        self.y_min = y_min
        self.y_max = y_max
        self.x_min = x_min
        self.x_max = x_max
        self.x_size = self.x_max - self.x_min
        self.y_size = self.y_max - self.y_min
        self.y_pixel_scale = y_pixel_scale
        self.x_pixel_scale = x_pixel_scale
        self.x_start = self.x_min - self.x_pixel_scale/2.0
        self.y_start = self.y_min - self.y_pixel_scale/2.0

    def new_from_x_and_y_scale(self, x_scale, y_scale):
        return InterpolationGeometry(y_min=self.y_min * y_scale, y_max=self.y_max * y_scale,
                                     x_min=self.x_min * x_scale, x_max=self.x_max * x_scale,
                                     y_pixel_scale=self.y_pixel_scale*y_scale,
                                     x_pixel_scale=self.x_pixel_scale*x_scale)


class InterpolationScheme(object):

    def __init__(self, shape, image_coords, image_pixel_scale):
        """An interpolation scheme which enables efficient deflection angle computation. This works by computing a \
        sub-set of deflection angles on a uniform regular grid and bilinearly interpolating these values to determine \
        the image-grid and sub-grid deflection angle values.

        This class describes the interpolation scheme that is used to do this, for example the shape of the \
        interpolation-grid and the neighbors of every pixel on the grid in the 4 directions required for bilinear \
        interpolation.

        The interpolation scheme can be performed on image and sub grids that are non-uniform (e.g. after tracing them \
        from the image-plane to a source-plane). The x_pixel and y_pixel bounds of the new grids are used to scale the \
        interpolation scheme to the new plane.

        Parameters
        -----------
        shape : (int, int)
            The shape of the interpolation schemes interpolation-grid.
        image_coords : ndarray
            The image-plane coordinates of each pixel on the interpolation-grid.
        """
        def bottom_right_neighbors():
            """For each pixel on the deflection-interpolation grid, compute pixels directly neighboring each pixel \
            to their right, downwards and down-right.

            These are the pixels bilinear interpolation will be performed using if a deflection angle lands in the \
            bottom-right quadrant of a pixel."""

            down_right_neighbors = -1*np.ones((self.pixels, 3), dtype='int')

            for y_pixel in range(self.shape[0]):
                for x_pixel in range(self.shape[1]):

                    pixel_index = y_pixel * self.shape[1] + x_pixel

                    if x_pixel < self.shape[1]-1:
                        down_right_neighbors[pixel_index, 0] = pixel_index+1

                    if y_pixel < self.shape[0]-1:
                        down_right_neighbors[pixel_index, 1] = pixel_index+self.shape[1]

                    if x_pixel < self.shape[1]-1 and y_pixel < self.shape[0]-1:
                        down_right_neighbors[pixel_index, 2] = pixel_index+self.shape[1]+1

            return down_right_neighbors

        def bottom_left_neighbors():
            """For each pixel on the deflection-interpolation grid, compute pixels directly neighboring each pixel \
            to their left, downwards and down-left.

            These are the pixels bilinear interpolation will be performed using if a deflection angle lands in the \
            bottom-left quadrant of a pixel."""

            down_left_neighbors = -1*np.ones((self.pixels, 3), dtype='int')

            for y_pixel in range(self.shape[0]):
                for x_pixel in range(self.shape[1]):

                    pixel_index = y_pixel * self.shape[1] + x_pixel

                    if x_pixel > 0:
                        down_left_neighbors[pixel_index, 0] = pixel_index-1

                    if x_pixel > 0 and y_pixel < self.shape[0]-1:
                        down_left_neighbors[pixel_index, 1] = pixel_index+self.shape[1]-1

                    if y_pixel < self.shape[0]-1:
                        down_left_neighbors[pixel_index, 2] = pixel_index+self.shape[1]

            return down_left_neighbors

        def top_right_neighbors():
            """For each pixel on the deflection-interpolation grid, compute pixels directly neighboring each pixel \
            to their right, upwards and up-right.

            These are the pixels bilinear interpolation will be performed using if a deflection angle lands in the \
            top-right quadrant of a pixel."""
            up_right_neighbors = -1*np.ones((self.pixels, 3), dtype='int')

            for y_pixel in range(self.shape[0]):
                for x_pixel in range(self.shape[1]):

                    pixel_index = y_pixel * self.shape[1] + x_pixel

                    if y_pixel > 0:
                        up_right_neighbors[pixel_index, 0] = pixel_index-self.shape[1]

                    if x_pixel < self.shape[1]-1 and y_pixel > 0:
                        up_right_neighbors[pixel_index, 1] = pixel_index-self.shape[1]+1

                    if x_pixel < self.shape[1]-1:
                        up_right_neighbors[pixel_index, 2] = pixel_index+1

            return up_right_neighbors

        def top_left_neighbors():
            """For each pixel on the deflection-interpolation grid, compute pixels directly neighboring each pixel \
            to their left, upwards and up-left.

            These are the pixels bilinear interpolation will be performed using if a deflection angle lands in the \
            top-left quadrant of a pixel."""
            up_left_neighbors = -1*np.ones((self.pixels, 3), dtype='int')

            for y_pixel in range(self.shape[0]):
                for x_pixel in range(self.shape[1]):

                    pixel_index = y_pixel * self.shape[1] + x_pixel

                    if x_pixel > 0 and y_pixel > 0:
                        up_left_neighbors[pixel_index, 0] = pixel_index-self.shape[1]-1

                    if y_pixel > 0:
                        up_left_neighbors[pixel_index, 1] = pixel_index-self.shape[1]

                    if x_pixel > 0:
                        up_left_neighbors[pixel_index, 2] = pixel_index-1


            return up_left_neighbors

        self.image_coords = image_coords
        self.image_pixel_scale = image_pixel_scale

        self.geometry = InterpolationGeometry(y_min=np.min(image_coords[:,1]), y_max=np.max(image_coords[:,1]),
                                              x_min=np.min(image_coords[:,0]), x_max=np.max(image_coords[:,0]),
                                              y_pixel_scale=image_pixel_scale, x_pixel_scale=image_pixel_scale)

        self.shape = shape
        self.pixels = self.shape[0] * self.shape[1]

        self.bottom_right_neighbors = bottom_right_neighbors()
        self.bottom_left_neighbors = bottom_left_neighbors()
        self.top_right_neighbors = top_right_neighbors()
        self.top_left_neighbors = top_left_neighbors()

    @classmethod
    def from_mask(cls, mask, shape):
        """Determine the interpolation scheme from an image-mask. This uses the x / y_pixel bounds of the mask to setup the \
        grid 'over' the mask, padded by the pixel-scale to ensure edge pixels have their deflection angles interpolated \
        correctly.

        Parameters
        -----------
        mask: msk.Mask
            The mask the interpolation scheme is generated based on.
        shape : (int, int)
            The shape of the interpolation schemes interpolation-grid.
        """
        coordinate_grid = mask.coordinate_grid

        x_max = np.max(coordinate_grid[:,0]) + mask.pixel_scale
        x_min = np.min(coordinate_grid[:,0]) - mask.pixel_scale
        y_max = np.max(coordinate_grid[:,1]) + mask.pixel_scale
        y_min = np.min(coordinate_grid[:,1]) - mask.pixel_scale

        image_coords = np.zeros((shape[0] * shape[1], 2))

        for y_pixel in range(shape[0]):
            for x_pixel in range(shape[1]):

                pixel_index = y_pixel * shape[1] + x_pixel

                image_coords[pixel_index,1] = x_min + 2.0*(x_pixel/(shape[1]-1))*x_max
                image_coords[pixel_index,0] = y_min + 2.0*(y_pixel/(shape[0]-1))*y_max

        return InterpolationScheme(shape, image_coords, image_pixel_scale=mask.pixel_scale)

    def interpolation_coordinates_from_sizes(self, new_x_size, new_y_size):
        """Setup a set of interpolation coordinates, which represent a uniform-grid of coordinates which will be used \
        for deflection angle interpolation. These coordinates use the interpolation-scheme."""

        x_scale = new_x_size / self.geometry.x_size
        y_scale = new_y_size / self.geometry.y_size

        new_geometry = self.geometry.new_from_x_and_y_scale(x_scale, y_scale)

        interp_coords = np.zeros((self.pixels, 2), dtype='float64')

        interp_coords[:,0] = self.image_coords[:,0] * x_scale
        interp_coords[:,1] = self.image_coords[:,1] * y_scale

        return InterpolationCoordinates(array=interp_coords, geometry=new_geometry, scheme=self)


class InterpolationCoordinates(np.ndarray):

    def __new__(cls, array, geometry, scheme, *args, **kwargs):
        coords = np.array(array).view(cls)
        coords.geometry = geometry
        coords.scheme = scheme
        return coords

    def apply_function(self, func):
        return InterpolationDeflections(func(self), self, self.geometry, self.scheme)

    def interpolation_deflections_from_coordinates_and_galaxies(self, galaxies):

        def calculate_deflections(grid):
            return sum(map(lambda galaxy: galaxy.deflections_from_grid(grid), galaxies))

        return self.apply_function(calculate_deflections)


class InterpolationDeflections(np.ndarray):

    def __new__(cls, array, coords, geometry, scheme, *args, **kwargs):
        defls = np.array(array).view(cls)
        defls.interp_coords = coords
        defls.geometry = geometry
        defls.scheme = scheme
        return defls

    def grid_to_interp_from_grid(self, grid):

        y_pixels = np.floor((grid[:,1] - self.geometry.y_start) / self.geometry.y_pixel_scale)
        x_pixels = np.floor((grid[:,0] - self.geometry.x_start) / self.geometry.x_pixel_scale)

        return np.floor((x_pixels * self.scheme.shape[1]) + y_pixels)

    def interpolate_values_from_grid(self, grid):

        grid_to_interp = self.grid_to_interp_from_grid(grid)

        interpolated = np.zeros(grid.shape)

        for i in range(grid.shape[0]):

            interp_index = grid_to_interp[i]

            print(interp_index)
            print(self.interp_coords[interp_index, 0])

            if grid[i, 0] < self.interp_coords[interp_index, 0]:
                if grid[i, 1] < self.interp_coords[interp_index, 1]:
                    interpolated[i,0] = self.interpolate_in_top_left_of_pixel(grid[i, 0], grid[i, 1], interp_index,
                                                                              self[:, 0])
                    interpolated[i,1] = self.interpolate_in_top_left_of_pixel(grid[i, 0], grid[i, 1], interp_index,
                                                                              self[:, 1])

        return interpolated

    def interpolate_in_top_left_of_pixel(self, x, y, interp_index, deflections):

        bottom_right_index = interp_index
        top_left_index = self.scheme.top_left_neighbors[interp_index, 0]
        top_right_index = self.scheme.top_left_neighbors[interp_index, 1]
        bottom_left_index = self.scheme.top_left_neighbors[interp_index, 2]
        x0 = self.interp_coords[top_left_index, 0]
        x1 = self.interp_coords[top_right_index, 0]
        y0 = self.interp_coords[top_left_index, 1]
        y1 = self.interp_coords[bottom_left_index, 1]
        weight0 = ((x1-x)/(x1-x0))*deflections[bottom_left_index] + ((x-x0)/(x1-x0))*deflections[bottom_right_index]
        weight1 = ((x1-x)/(x1-x0))*deflections[bottom_left_index] + ((x-x0)/(x1-x0))*deflections[top_right_index]
        return((y1-y)/(y1-y0))*weight0 + ((y-y0)/(y1-y0))*weight1


from autolens import config
import requests


class TestCase(object):
    def test_is_config(self):
        assert config.is_config()

    def test_is_hosted(self):
        with requests.get(config.CONFIG_URL) as response:
            assert response.status_code == 200
